--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Animation - A sequence of frames used by a Costume.

You shouldn't normally need to use this API directly as the Costume
automatically loads any animations it finds.
--]]


require "lib.klutils.cabin"
require "lib.klutils.oops"

-- Private fields go here
Animation = BaseObj:makeClass
{
	_animspeed = 0.25,
	_numFrames = 0,
	_loopMethod = 'loop',
	_loopPoint = 1
}

-- Loads an animation out of the given costume folder and populates the list of
--  frames and sets any properties.
function Animation:init(costumeName, animationName)
	if not costumeName then
		return
	end
	
	-- Root folder
	local filepath = "data/costumes/" .. costumeName
	
	-- Folder where we will get the costumes from. If no animation is specified
	--  we just use the root folder.
	if animationName then
		filepath = filepath .. "/" .. animationName
	end
	
	if not love.filesystem.isDirectory(filepath) then
		cabin.error("Could not open animation \"" .. costumeName .. "/" .. animationName .. "\". Does not exist.")
	end
	
	local fileTable = love.filesystem.enumerate(filepath)
	
	local index = 1
	self._frames = {}
	
	for i, v in ipairs(fileTable) do
		local filename = filepath .. "/" .. v
		
		if v ~= "properties.lua" and love.filesystem.isFile(filename)
					and not love.filesystem.isDirectory(filename) then
			
			self._frames[index] = love.graphics.newImage(filename)
			if self._frames[index] then
				index = index + 1
			end
		end
	end
	
	self._numFrames = index - 1
	
	-- Animation properties
	local properties = propertyfile.get(filepath)
	
	if not properties then
		cabin.warning("Couldn't load property file for costume \"" .. costumeName
								.. "\", animation \"" .. animationName .. "\"")
	else
		self._animspeed = tonumber(properties.speed) or 0.25
		self._loopMethod = properties.loop_method
						and tostring(properties.loop_method) or 'loop'
		self._loopPoint = tonumber(properties.loop_point) or 1
	end
end

-- Getters --

function Animation:getSpeed()
	return self._animspeed
end

function Animation:getFramesCount()
	return self._numFrames
end

function Animation:getFrame(index)
	if self._frames then
		return self._frames[index]
	end
end

function Animation:getLoopMethod()
	return self._loopMethod
end

function Animation:getLoopPoint()
	return self._loopPoint
end

