--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Entity - a moving object that populates a Scene.

Generally, you'll superclass Entity and provide your own methods for init()
and think(). For example, you might make a call to setCostume() inside
your init() method, and use think() to script the behaviors for your entity.
--]]


require "lib.klutils.cabin"
require "lib.klutils.objectlist"
require "lib.klutils.oops"

-- Make math operations faster, code cleaner
local floor = math.floor

-- Private fields go here
Entity = ListedObj:makeClass
{
	-- TODO: Costume cache should be in costume.lua, not here.
	_static = {
		_costumeCache = {},
		_costumeStrongRefs = {}
	},
	_x = 0,
	_y = 0,
	_w = 1,
	_h = 1,
	_xVel = 0,
	_yVel = 0,
	_xAcc = 0,
	_yAcc = 0,
	_angle = 0,
	_hflip = false,
	_vflip = false,
	_color = {255, 255, 255, 255},
	_costumeAnimation = "idle",
	_costumeFrame = 1,
	_costumeTimer = 0,
	_visible = true
}

-- Cleaning the cache --
-- In order to provide fast loading for resources previously loaded while the
--  scene is running, we have a cache that simply holds a reference that any
--  new objects can point to. But when we switch to a new scene, the resource
--  is still loaded even if we don't need it!
-- Thus we have a costume cache which uses weak references. We have a second
--  table which includes strong references to the same resources - we simply
--  clear it when we want to be able to empty the cache.

local __entityCache_mt = {}
setmetatable(Entity._static._costumeCache, __entityCache_mt)
__entityCache_mt.__mode = "v"

function Entity:init(x, y)
	
	if x and y then
		self._x = x
		self._y = y
	end
end

function cleanCostumeCache()
	Entity._static._costumeStrongRefs = {}
end

-- Getters --

function Entity:getCostume()
	return self._costume
end

function Entity:getAnimationName()
	return self._costumeAnimation
end

function Entity:getFrameNumber()
	return self._costumeFrame
end

function Entity:getPosition()
	return self._x, self._y
end

function Entity:getVelocity()
	return self._xVel, self._yVel
end

function Entity:getAcceleration()
	return self._xAcc, self._yAcc
end

function Entity:getAngle()
	return self._angle
end

function Entity:getScale()
	return self._w, self._h
end

function Entity:getFlip()
	return self._hflip, self._vflip
end

function Entity:getColor()
	return self._color[1], self._color[2], self._color[3], self._color[4]
end

function Entity:isVisible()
	return self._visible
end

-- Setters --

function Entity:setCostume(newCostume)
	if newCostume then
		if self._static._costumeCache[newCostume] then
			self._costume = self._static._costumeCache[newCostume]
		else
			local costume = Costume:new(newCostume)
			
			self._static._costumeCache[newCostume] = costume
			self._static._costumeStrongRefs[newCostume] = costume
			self._costume = costume
		end
	end
end

function Entity:setAnimation(name)
	if not name then
		return
	end
	
	if self._costume:hasAnimation(name) then
		self._costumeAnimation = name
		self._costumeFrame = 1
		self._costumeTimer = 0
	else
		cabin.warning("Costume \"" .. self._costume:getName() .. "\" does not have animation \"" .. name .. "\"" )
	end
end

function Entity:setPosition(x, y)
	if x and y then
		self._x = x
		self._y = y
	end
end

function Entity:setVelocity(xVel, yVel)
	if xVel and yVel then
		self._xVel = xVel
		self._yVel = yVel
	end
end

function Entity:setAcceleration(xAcc, yAcc)
	if xAcc and yAcc then
		self._xAcc = xAcc
		self._yAcc = yAcc
	end
end

function Entity:setAngle(angle)
	if angle then
		self._angle = angle
	end
end

function Entity:setScale(w, h)
	if w and h then
		self._w = w
		self._h = h
	end
end

function Entity:setFlip(hflip, vflip)
	self._hflip = (hflip == true)
	self._vflip = (vflip == true)
end

function Entity:setColor(r, g, b, a)
	if r and g and b then
		self._color = {r, g, b, a or 255}
	end
end

function Entity:setVisible(isTrue)
	self._visible = (isTrue == true)
end

-- Extra Methods --

function Entity:think()
	
end

function Entity:costumeStep()
	if not self._costume then
		return
	end
	
	local animation = self._costume:getAnimation(self._costumeAnimation)
	
	if not animation then
		return
	end
	
	self._costumeTimer = self._costumeTimer + animation:getSpeed()
	
	
	while self._costumeTimer >= 1.0 do
		self._costumeTimer = self._costumeTimer - 1
		self._costumeFrame = self._costumeFrame + 1
		
		if self._costumeFrame > animation:getFramesCount() then
			local loopMethod = animation:getLoopMethod()
			
			if loopMethod == "loop" then
				self._costumeFrame = animation:getLoopPoint()
			elseif loopMethod == "freeze" then
				self._costumeFrame = animation:getFramesCount()
			elseif loopMethod == "restore" then
				self:setAnimation("idle")
			end
		end
	end
	
end

function Entity:move()
	self._x = self._x + self._xVel
	self._y = self._y + self._yVel
	self._xVel = self._xVel + self._xAcc
	self._yVel = self._yVel + self._yAcc
end

function Entity:getImage()
	if not self._costume then
		return
	end
	
	local animation = self._costume:getAnimation(self._costumeAnimation)
	
	if animation then
		return animation:getFrame(self._costumeFrame)
	end
end

function Entity:draw()
	local image = self:getImage()

	if image then
		love.graphics.setColor(self._color)
		love.graphics.draw(image, floor(self._x + 0.5), floor(self._y + 0.5), self._angle,
				(self._hflip == true) and -self._w or self._w,
				(self._vflip == true) and -self._h or self._h,
				floor(image:getWidth() / 2), floor(image:getHeight() / 2))
	end
end

function Entity:destroy()
	local scene = self:getParent()
	
	if scene and scene.removeEntity then
		scene:removeEntity(self)
	end
end
