--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Gameloop - Provides facilities to control the gameloop.

You'll use these when you want to adjust things like the current screen scaling
mechanism or switch to a different Scene within the game.
--]]


-- Private variables
local _priv = {
	_deltaqueue = 0,
	_currentScene = nil,
	_pendingScene = nil,
	_sceneIndex = 0,
	_sceneCount = 0,
	_sceneList = {},
	_sceneNames = {},
	_enableDebug = false,
	_enableSleep = false,
	_framerate = 60.0,
	_droppedFrames = 0,
	_maxSkip = 5,
	_resolutionType = 'fixed',
	_referenceWidth = -1,
	_referenceHeight = -1,

	_timingTable = nil,
	_lastTime = 0,
	_debugFont = nil
}

gameloop = {}

function _priv._debugTiming(name)
	if not _priv._timingTable then
		_priv._timingTable = {}
		_priv._lastTime = love.timer.getMicroTime()
	end
	
	local currTime = love.timer.getMicroTime()
	
	_priv._timingTable[name] = currTime - _priv._lastTime
	_priv._lastTime = currTime
end

function _priv._getDebugTiming(name)
	if _priv._timingTable and _priv._timingTable[name] then
		return _priv._timingTable[name]
	end
	
	return 0
end

function _priv._drawShadowedText(string, x, y)
	local r, g, b, a = love.graphics.getColor()
	
	love.graphics.setColor(0, 0, 0, 255)
	love.graphics.print(string, x + 1, y + 1)
	love.graphics.setColor(r, g, b, a)
	love.graphics.print(string, x, y)
end

function _priv._drawDebug()
	local pos = 0;
	
	for i, timeName in ipairs({ "Wait", "Other", "Present", "Draw", "Logic" }) do
		_priv._drawShadowedText(timeName .. ": "
			.. math.floor(_priv._getDebugTiming(timeName) * 10000) / 10
			.. " ms", 2, love.graphics.getHeight() - 14 - pos)
		pos = pos + 10
	end
end

function gameloop.getDebugEnabled()
	return (_priv._enableDebug == true)
end

function gameloop.setDebug(isTrue)
	_priv._enableDebug = (isTrue == true)
end

function gameloop.setSleep(isTrue)
	_priv._enableSleep = (isTrue == true)
end

function gameloop.setFramerate(number)
	_priv._framerate = number
end

function gameloop.setResolutionType(resType, width, height)
	if type(resType) == 'string' then
		_priv._resolutionType = resType
	end
	if width and height then
		_priv._referenceWidth = width
		_priv._referenceHeight = height
	end
end

function gameloop.setScene(sceneObj)
	_priv._sceneIndex = 0
	_priv._pendingScene = sceneObj
end

function gameloop.addScene(sceneClass, name)
	if sceneClass and name then
		_priv._sceneCount = _priv._sceneCount + 1
		_priv._sceneList[_priv._sceneCount] = sceneClass
		_priv._sceneNames[name] = _priv._sceneCount
	end
end

function gameloop.firstScene()
	local scene = _priv._sceneList[1]
	
	if scene then
		_priv._pendingScene = scene:new()
		_priv._sceneIndex = 1
	end
end

function gameloop.nextScene()
	if _priv._sceneIndex + 1 <= _priv._sceneCount then
		_priv._sceneIndex = _priv._sceneIndex + 1
		_priv._pendingScene = _priv._sceneList[_priv._sceneIndex]:new()
	end
end

function gameloop.previousScene()
	if _priv._sceneIndex - 1 > 0 then
		_priv._sceneIndex = _priv._sceneIndex - 1
		_priv._pendingScene = _priv._sceneList[_priv._sceneIndex]:new()
	end
end

function gameloop.setSceneByName(name)
	local index = _priv._sceneNames[name]
	
	if index then
		local scene = _priv._sceneList[index]
		if scene then
			_priv._pendingScene = scene:new()
			_priv._sceneIndex = index
		end
	end
end

function love.keypressed(key, unicode)
	if key == "f2" then
		-- Restart game
		gameloop.firstScene()
	elseif key == "f4" and love.keyboard.isDown("lalt")
						or love.keyboard.isDown("ralt") then
		-- Quit even when the user is running a stupid version of SDL that
		--  blocks outside keyboard shortcuts
		love.event.push('quit')
	elseif key == "f12" then
		-- Take screenshot
		local index = 1
		local filename = "screenshot1.png"
		
		while love.filesystem.exists(filename) do
			index = index + 1
			filename = "screenshot" .. index .. ".png"
		end
		
		local screenshot = love.graphics.newScreenshot()
		screenshot:encode(filename, 'png')
	end
end

function love.update(dt)
	_priv._deltaqueue = _priv._deltaqueue + dt
	
	-- Switch to a different scene?
	if _priv._currentScene ~= _priv._pendingScene then
		_priv._currentScene = _priv._pendingScene
		
		-- Clean up any assets that are no longer being used.
		-- Seriously, this is like one of the few game making tool runtimes to
		--  actually do something like this. At least _I_ care about not
		--  causing the end user's machine to run out of memory due to assets
		--  never getting unloaded ever.
		cleanCostumeCache()
		cleanTileCache()
		cleanSoundCache()
		collectgarbage()
	end
	
	_priv._droppedFrames = -1
	local framesLeft = true
	-- Skip frames, but don't skip more than the max
	while framesLeft == true do
		_priv._deltaqueue = _priv._deltaqueue - (1.0 / _priv._framerate)
		
		if _priv._droppedFrames < _priv._maxSkip then
			if _priv._currentScene then
				_priv._currentScene:prestep()
				
				input.poll()
				
				_priv._currentScene:step()
				_priv._currentScene:poststep()
			end
			
			_priv._droppedFrames = _priv._droppedFrames + 1
		end
		
		framesLeft = (_priv._deltaqueue >= 1.0 / _priv._framerate)
	end
end

function love.draw()
	
	if not _priv._debugFont then
		_priv._debugFont = love.graphics.newFont("lib/klua/data/debug_font.ttf", 8)
	end
	
	if not _priv._currentScene then
		love.graphics.setBackgroundColor(0, 0, 0)
		love.graphics.setFont(_priv._debugFont)
		love.graphics.setColor(255, 255, 255, 255)
		love.graphics.print("No scene loaded.\nDid you remember to call\n    gameloop.addScene()\n    or gameloop.setScene()\n    within setup()?", 20, 20)
		return
	end
	
	local width, height = love.graphics.getWidth(), love.graphics.getHeight()
	if _priv._resolutionType == 'fixed' then
		_priv._currentScene:draw(0, 0, width, height, width, height)
	elseif _priv._resolutionType == 'independent' then
		local viewWidth = (_priv._referenceWidth > -1) and _priv._referenceWidth or width
		local viewHeight = (_priv._referenceHeight > -1) and _priv._referenceHeight or height
		_priv._currentScene:draw(0, 0, viewWidth, viewHeight, width, height)
	end
	
	-- Debug info
	love.graphics.setFont(_priv._debugFont)
	love.graphics.setColor(255, 255, 255, 255)
	
	if _priv._enableDebug == true then
		if _priv._sceneIndex > 0 then
			_priv._drawShadowedText("Scene " .. _priv._sceneIndex .. ": " .. _priv._currentScene:getName(), 2, 0)
		else
			_priv._drawShadowedText("Scene: " .. _priv._currentScene:getName(), 2, 0)
		end
		_priv._drawShadowedText("Entities: " .. _priv._currentScene:getEntityCount(), 2, 10)
		_priv._drawShadowedText("Tics: " .. _priv._currentScene:getTics(), 2, 20)
		_priv._drawShadowedText("FPS: " .. love.timer.getFPS(), 2, 30)
		_priv._drawShadowedText("Dropped: " .. _priv._droppedFrames, 62, 30)
		_priv._drawShadowedText("Lua: " .. math.floor(collectgarbage("count")) .. " kB", 2, 40)
		_priv._drawDebug()

	end
end

-- Overriding the default love.run here

function love.run()
	
	math.randomseed(os.time())
	math.random() math.random()
	
	if love.load then love.load(arg) end
	
	local dt = 0
	local sleepBegin = 0
	
	-- Main loop time.
	while true do
		-- Process events.
		if love.event then
			love.event.pump()
			
			for e,a,b,c,d in love.event.poll() do
				if e == "quit" then
					if not love.quit or not love.quit() then
						if love.audio then
							love.audio.stop()
						end
						return
					end
				end
				
				love.handlers[e](a,b,c,d)
			end
		end
		
		_priv._debugTiming("Other")
		
		-- Better sleep than the default, horribly broken love.run()
		-- Also get the actual delta...
		if love.timer then
			local toSleep = 1.0 / _priv._framerate
			
			local sleepTime = toSleep - (love.timer.getMicroTime() - sleepBegin)
			
			while sleepTime > 0 do
				if _priv._enableSleep == true and sleepTime > .0002 then
					love.timer.sleep(.0001)
				end
				sleepTime = toSleep - (love.timer.getMicroTime() - sleepBegin)
			end
			
			sleepBegin = love.timer.getMicroTime()
			
			love.timer.step()
			dt = love.timer.getDelta()
		end
		
		_priv._debugTiming("Wait")
		
		love.update(dt)
		
		_priv._debugTiming("Logic")
		
		if love.graphics then
			love.graphics.clear()
			love.draw()
		end
		
		_priv._debugTiming("Draw")
		
		if love.graphics then
			love.graphics.present()
		end
		
		_priv._debugTiming("Present")
		
	end
end
