--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Input - Helper functions to bind keys to different in-game controls, as well
as test for whether these controls are [not] held down, and for how long.
--]]


local _priv = {
	_keyList = {},
	_keyState = {}
}

input = {}

function input.setDefaults()
	_priv._keyList = {
		["left"] = {"left"},
		["right"] = {"right"},
		["up"] = {"up"},
		["down"] = {"down"},
		["fire1"] = {"lshift", "rshift"},
		["fire2"] = {"lctrl", "rctrl"},
		["fire3"] = {" "},
		["fire4"] = {"a"},
		["fire5"] = {"s"},
		["fire6"] = {"d"}
	}
end

function input.setControl(controlName, keys)
	if type(keys) == "table" then
		_priv._keyList[controlName] = keys
	elseif type(keys) == "string" then
		-- Make it easier to supply a single key binding
		_priv._keyList[controlName] = { keys }
	end
end

function input.held(controlName)
	return (input.timeHeld(controlName) > 0)
end

function input.poll()
	for bind, keytable in pairs(_priv._keyList) do
		-- Create the entry in the state table if it's not there yet
		if not _priv._keyState[bind] then
			_priv._keyState[bind] = 0
		end
		local pressed = false
		
		for index, key in ipairs(keytable) do
			if love.keyboard.isDown(key) then
				pressed = true
				break
			end
		end
		
		if pressed == true then
			if _priv._keyState[bind] > 0 then
				_priv._keyState[bind] = _priv._keyState[bind] + 1
			else
				_priv._keyState[bind] = 1
			end
		else
			if _priv._keyState[bind] < 0 then
				_priv._keyState[bind] = _priv._keyState[bind] - 1
			else
				_priv._keyState[bind] = -1
			end
		end
	end
end

function input.timeHeld(controlName)
	if not _priv._keyState[controlName] or _priv._keyState[controlName] < 0 then
		return 0
	end
	
	return _priv._keyState[controlName]
end

function input.timeReleased(controlName)
	if not _priv._keyState[controlName] or _priv._keyState[controlName] > 0 then
		return 0
	end
	
	return 0 - _priv._keyState[controlName]
end

function input.downOnce(controlName)
	return (input.timeHeld(controlName) == 1)
end

function input.upOnce(controlName)
	return (input.timeReleased(controlName) == 1)
end

function input.autoRepeat(controlName, initialDelay, repeatRate)
	local time = input.timeHeld(controlName)
	return (time and (time == 1 or (time >= initialDelay
				and (time - initialDelay) % repeatRate == 0)))
end
