--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Scene - A map, screen, or level where all the action takes place. A scene
includes a tilegrid and any number of entities (as long as that number is below
maxObjs), and can be used for things like title screens, main menus, game
levels, etc.

TODO: Soon Scene will contain layers which will in turn hold the tilegrids and
entities...
--]]


require "lib.klutils.cabin"
require "lib.klutils.objectlist"
require "lib.klutils.oops"

-- Make math operations faster, code cleaner
local floor, ceil, min, max = math.floor, math.ceil, math.min, math.max

-- Hardcoded limit. Technically the scene can hold as many objects as can be
--  represented by a double-precision float but this is mainly for the sake
--  of sanity, especially since things like deletion costs start getting heavy
--  with this many objects.
local _maxObjs = 20000

-- Private fields go here
Scene = BaseObj:makeClass
{
	_name = "Untitled Scene",
	_width = 800,
	_height = 600,
	_scrollX = 0,
	_scrollY = 0,
	_bgColor = {120, 208, 240},
	_tics = 0,
	_objectList = nil,
	_tilegrid = nil
}

function Scene:init(sceneName)
	self._objectList = ObjectList:new(self)
	
	if not sceneName then
		return
	end
	
		-- Root folder
	local filepath = "data/scenes/" .. sceneName
	
	if not love.filesystem.isDirectory(filepath) then
		cabin.error("Could not open scene \"" .. sceneName .. "\". Does not exist.")
	end
	
	-- Scene properties/data
	local properties = propertyfile.get(filepath)
	
	if not properties then
		cabin.error("Oops: Couldn't load property file for scene \"" .. sceneName .. "\"")
	else
		local tilegrid = Tilegrid:new(tostring(properties.tileset))
		tilegrid:setTilesFromTable(properties.tiles)
		
		self:setTilegrid(tilegrid)
	end
end

-- Getters --

function Scene:getName()
	return self._name
end

function Scene:getAreaSize()
	return self._width, self._height
end

function Scene:getBackgroundColor()
	return self._bgColor[1], self._bgColor[2], self._bgColor[3]
end

function Scene:getEntityCount()
	return self._objectList:count()
end

function Scene:getTilegrid()
	return self._tilegrid
end

function Scene:getScrollVector()
	return self._scrollX, self._scrollY
end

-- Setters --

function Scene:setName(name)
	if name then
		self._name = name
	end
end

function Scene:setAreaSize(width, height)
	if width and height then
		-- Don't accept negative values for the size
		width = (width >= 0) and width or 0
		height = (height >= 0) and height or 0
		self._width = width
		self._height = height
	end
end

function Scene:setBackgroundColor(red, green, blue)
	if red and green and blue then
		self._bgColor = {red, green, blue}
	end
end

function Scene:setTilegrid(tilegridObj)
	if tilegridObj then
		self._tilegrid = tilegridObj
		
		local tileset = self._tilegrid:getTileset()
		
		if tileset then
			-- Automatically change the size of the scene to fit.
			local w, h = self._tilegrid:getSize()
			self._width = w * tileset:getTileSize()
			self._height = h * tileset:getTileSize()
		end
	end
end

function Scene:centerScrollVector(x, y)
	if x and y then
		self._scrollX = x
		self._scrollY = y
	end
end

-- Extra Methods --

function Scene:addEntity(entityObj)
	if self._objectList:count() >= _maxObjs then
		return false
	end
	
	return self._objectList:insertAndOwn(entityObj)
end

function Scene:removeEntity(entityObj)
	if not entityObj then
		return
	end
	
	self._objectList:release(entityObj)
end

function Scene:prestep()
	
end

function Scene:step()
	self._tics = self._tics + 1
	
	self._objectList:iterate(_processEntity)
	
	for i, v in self._objectList:iterate() do
		v:think()
		v:move()
		if v:getCostume() then
			v:costumeStep()
		end
	end
	
	self._objectList:sync()
end

function Scene:poststep()
	
end

function Scene:draw(x, y, viewW, viewH, screenW, screenH)
	
	if self._width < 0 or self._height < 0 then
		return
	end
	
	-- Push previous state to stack
	love.graphics.push()
	
	-- Pick the smaller of the two viewport-to-screen ratios
	local scale = min(screenW / viewW, screenH / viewH)
	
	-- Is the scene actually smaller than the view? If so, readjust the coords.
	viewW = min(viewW, self._width)
	viewH = min(viewH, self._height)
	
	-- After the scale, set the new screen coordinates to provide the black bars
	x = x + ceil((screenW - viewW * scale) / 2)
	y = y + ceil((screenH - viewH * scale) / 2)
	screenW = viewW * scale
	screenH = viewH * scale
	
	-- Scrolling offset, taking point-of-no-scroll into account.
	local offsetX = min(max(self._scrollX, viewW / 2),
									self._width - viewW / 2)
	local offsetY = min(max(self._scrollY, viewH / 2),
									self._height - viewH / 2)
	
	love.graphics.setScissor(max(x, 0), max(y, 0),
							min(floor(screenW), self._width * scale),
							min(floor(screenH), self._height * scale))
	
	love.graphics.setBackgroundColor(self._bgColor)
	love.graphics.clear()
	love.graphics.setBackgroundColor(0, 0, 0)
	
	-- Translate to the center of the screen so that we scale from the center
	love.graphics.translate(floor(screenW / 2) + x, floor(screenH / 2) + y)
	love.graphics.scale(scale, scale)
	-- Translate to the scroll vector
	love.graphics.translate(floor(-offsetX + 0.5), floor(-offsetY + 0.5))
	
	if self._tilegrid then
		self._tilegrid:draw(offsetX - viewW / 2, offsetY - viewH / 2,
				offsetX + viewW / 2, offsetY + viewH / 2)
	end
	
	for i, v in self._objectList:iterate() do
		if v:isVisible() then
			v:draw()
		end
	end
	
	-- Restore state
	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.setScissor()
	love.graphics.pop()
end

function Scene:getTics()
	return self._tics
end
