 --[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Sound - A sound that can be played back. Sources are automatically managed
in the background.
--]]


--[[
NOTICE: love.audio should go die in a fire.

Previously, this system would call love.audio.newSource() every time a sound
was played. This is the _correct_ way to do things, but unfortunately due to
LOVE's inane memory management, 1MB of additional memory would be allocated
on every single call to newSource() and this memory would never get returned
to the operating system, even if the source variable was set to nil and the
garbage collector forced to collect unused resources.

Now we are pre-allocating Sources. This is completely WRONG and also wastes
excessive memory and the memory still doesn't get freed up if the sounds get
cleared from the cache, but at least it doesn't leak MORE memory every time a
sound gets played.

Once love.audio gets overhauled and is less incompetent, this will revert back
to the original implementation.
--]]

require "lib.klutils.cabin"
require "lib.klutils.oops"

-- Private fields go here
Sound = BaseObj:makeClass
{
	_static = {
		_soundCache = {},
		_soundStrongRefs = {}
	},
	_sounds = nil,
	_numSounds = 0
}

local __soundCache_mt = {}
setmetatable(Sound._static._soundCache, __soundCache_mt)
__soundCache_mt.__mode = "v"

function cleanSoundCache()
	Sound._static._soundStrongRefs = {}
end

function Sound:init(names, polyphony)
	if not names then
		return
	end
	
	polyphony = polyphony or 4
	
	local nameList = {}
	
	if type(names) == 'string' then
		nameList = { names }
	else
		nameList = names
	end
	
	self._sounds = {}
	
	for i, v in ipairs(nameList) do
		-- Check sounds in cache
		if self._static._soundCache[v] then
			self._sounds[i] = self._static._soundCache[v]
		else
			local soundTable = {}
			local filename = "data/sounds/" .. v
			
			if not love.filesystem.isFile(filename) then
				cabin.error("Could not open sound \"" .. v .. "\". Does not exist.")
			end
			
			soundTable.data = love.sound.newSoundData(filename)
			soundTable.numSources = polyphony
			
			for i = 1, polyphony do
				soundTable[i] = love.audio.newSource(soundTable.data, 'static')
			end
			
			self._sounds[i] = soundTable
			self._static._soundCache[v] = soundTable
		end
		
		self._numSounds = i
	end
end

function Sound:play()
	if self._sounds then
		local index = math.random(1, self._numSounds)
		local sound = self._sounds[index]
		local source = nil
		
		for i = 1, sound.numSources do
			if sound[i]:isStopped() then
				sound[i]:rewind()
				sound[i]:play()
				
				break
			end
		end
	end
end
