--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Tilegrid - A 2D array of tiles. A Tileset and a table of values gets fed in,
and the result is a tile backdrop that represents the game world.
--]]


require "lib.klutils.oops"

-- Make math operations faster, code cleaner
local floor, ceil,  max = math.floor, math.ceil, math.max

-- Hardcoded limit. Integrated graphics drivers really dislike large max sprite
--  batch sizes, and for performance reasons the sprite batch for drawing tiles is
--  only allocated once, so it can't be changed on the fly. For now we'll just set
--  the limit at a maximum of 15,000 tiles that can be drawn on the screen every
--  frame. 
local maxBatchTiles = 15000

-- Private fields go here
Tilegrid = BaseObj:makeClass
{
	_static = {},
	_tileset = nil,
	_width = 0,
	_height = 0,
	_tileValues = nil,
	_batch = nil
}

function Tilegrid:init(tilesetName, w, h)
	self._tileValues = {}
	
	self:setTileset(tilesetName)
	self:newGrid(w, h)
end

-- Getters --

function Tilegrid:getTileset()
	return self._tileset
end

function Tilegrid:getSize()
	return self._width, self._height
end

function Tilegrid:getTileValue(x, y)
	if self._tileValues and self._tileValues[y] then
		local value = self._tileValues[y][x]
		
		if value then
			return value
		end
	end
	
	return 0
end

-- Setters --

function Tilegrid:setTileset(tilesetName)
	if tilesetName then
		self._tileset = Tileset:new(tilesetName)
		
		local image = self._tileset:getImage()
		
		self._batch = love.graphics.newSpriteBatch(image, maxBatchTiles)
	end
end

function Tilegrid:setTileValue(x, y, value)
	if x and y and x >= 0 and y >= 0 and x < self._width and y < self._height then
		self._tileValues[y][x] = value
	end
end

function Tilegrid:setTilesFromTable(valueTable)
	local width = 0
	local y = 0

	self._tileValues = {}
	
	for k1, v1 in pairs(valueTable) do
		if type(v1) == 'table' then
			local x = 0
			
			self._tileValues[y] = {}
			
			for k2, v2 in pairs(v1) do
				if type(v2) == 'number' then
					self._tileValues[y][x] = v2
				end
				
				x = x + 1
			end
			
			-- Extend the horizontal bounds
			width = math.max(width, x)
			
			y = y + 1
		end
	end
	
	self._width = width
	self._height = y
end

-- Extra Methods --

function Tilegrid:newGrid(w, h)
	if w and h then
		self._tileValues = {}
		
		for y = 0, h - 1 do
			self._tileValues[y] = {}
			for x = 0, w - 1 do
				self._tileValues[y][x] = 0
			end
		end
		
		self._width = w
		self._height = h
	end
end

function Tilegrid:draw(fromX, fromY, toX, toY)
	if not self._tileset or not self._batch then
		return
	end
	
	local tileSize = self._tileset:getTileSize()
	
	fromX = fromX or 0
	fromY = fromY or 0
	toX = toX or self._width * tileSize
	toY = toY or self._height * tileSize
	
	self._batch:clear()
	
	for y = math.floor(fromY / tileSize), math.ceil(toY / tileSize) - 1 do
		for x = math.floor(fromX / tileSize), math.ceil(toX / tileSize) - 1 do
			local value = self:getTileValue(x, y)
			local quad = self._tileset:_getQuad(value)
			
			if quad then
				self._batch:addq(quad, x * tileSize, y * tileSize)
			end
		end
	end

	love.graphics.draw(self._batch)
	
end
