--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
Tileset - A bitmap sheet containing a series of tiles that make up a Tilegrid.
--]]


require "lib.klutils.cabin"
require "lib.klutils.oops"
require "lib.klua.tilegrid"

-- Private fields go here
Tileset = BaseObj:makeClass
{
	_static = {
		_tilesetCache = {},
		_tilesetStrongRefs = {}
	},
	_tilesetImage = nil,
	_tilesize = 32,
	_numTiles = 0
}

-- Cleaning the cache --
-- In order to provide fast loading for resources previously loaded while the
--  scene is running, we have a cache that simply holds a reference that any
--  new objects can point to. But when we switch to a new scene, the resource
--  is still loaded even if we don't need it!
-- Thus we have a tileset cache which uses weak references. We have a second
--  table which includes strong references to the same resources - we simply
--  clear it when we want to be able to empty the cache.

local __tilesetCache_mt = {}
setmetatable(Tileset._static._tilesetCache, __tilesetCache_mt)
__tilesetCache_mt.__mode = "v"

-- Loads a tileset image, padding the sides of each tile first so that there is
--  no bleeding at the edges of tiles when filtered and scaled.
local function _loadPaddedImage(filename, tilesize)
	local original = love.image.newImageData(filename)
	
	local width, height = original:getWidth(), original:getHeight()
	local newWidth = width + 2 * math.ceil(width / tilesize)
	local newHeight = height + 2 * math.ceil(height / tilesize)
	
	if newWidth > 2048 or newHeight > 2048 then
		cabin.error("Tileset image " .. filename .. " is too large.")
	end
	
	local padded = love.image.newImageData(newWidth, newHeight)
	
	for y = 0, math.floor(height / tilesize) - 1 do
		for x = 0, math.floor(width / tilesize) - 1 do
			local oldX, oldY = x * tilesize, y * tilesize
			local newX = x * (tilesize + 2) + 1
			local newY = y * (tilesize + 2) + 1
			
			-- Copy the actual tile over
			padded:paste(original, newX, newY, oldX, oldY, tilesize, tilesize)
			
			-- Repeat each edge by one pixel. N, S, W, E.
			padded:paste(original, newX, newY - 1,
						oldX, oldY, tilesize, 1)
			padded:paste(original, newX, newY + tilesize,
						oldX, oldY + tilesize - 1, tilesize, 1)
			padded:paste(original, newX - 1, newY,
						oldX, oldY, 1, tilesize)
			padded:paste(original, newX + tilesize, newY,
						oldX + tilesize - 1, oldY, 1, tilesize)
			
			-- Cover the four corners, too. NW, NE, SW, SE.
			padded:setPixel(newX - 1, newY - 1,
				original:getPixel(oldX, oldY))
			padded:setPixel(newX + tilesize, newY - 1,
				original:getPixel(oldX + tilesize - 1, oldY))
			padded:setPixel(newX - 1, newY + tilesize,
				original:getPixel(oldX, oldY + tilesize - 1))
			padded:setPixel(newX + tilesize, newY + tilesize,
				original:getPixel(oldX + tilesize - 1, oldY + tilesize - 1))
		end
	end
	
	return love.graphics.newImage(padded)
end

function Tileset:init(tilesetName)
	self._quadList = {}
	
	if not tilesetName then
		return
	end
	
	if self._static._tilesetCache[tilesetName] then
		local cached = self._static._tilesetCache[tilesetName]
		
		for k, v in pairs(cached) do
			self[k] = v
		end
	else
		local properties = propertyfile.get("data/tilesets/" .. tilesetName)
		
		if not properties then
			cabin.error("Couldn't load property file for tileset \"" .. tilesetName .. "\"")
		else
			local prop_tilesize = properties.tile_size
			
			if prop_tilesize then
				if prop_tilesize < 8 then
					cabin.error("For sanity reasons, tile sizes less than 8 pixels are not supported.")
					self._tilesize = 8
				else
					self._tilesize = prop_tilesize
				end
			end
			
			local filename = "data/tilesets/" .. tilesetName .. "/tiles.png"
			
			if not love.filesystem.isFile(filename) then
				cabin.error("Could not open tiles.png for tileset \"" .. tilesetName .. "\". Does not exist.")
			end
			
			self._tilesetImage = _loadPaddedImage(filename, self._tilesize)
			
			local width = self._tilesetImage:getWidth()
			local height = self._tilesetImage:getHeight()
			local newTilesize = self._tilesize + 2
			
			-- LOVErs will probably be wondering why I'm creating so many Quads
			--  here instead of simply using a single Quad and shifting its
			--  viewport. I tried both methods and as it turns out, it is
			--  faster to forego shifting the viewport - up to about 25% more
			--  performant.
			
			for y = 0, math.floor(height / newTilesize) - 1 do
				for x = 0, math.floor(width / newTilesize) - 1 do
					self._numTiles = self._numTiles + 1
					self._quadList[self._numTiles]
							= love.graphics.newQuad(x * newTilesize + 1,
													y * newTilesize + 1,
													self._tilesize,
													self._tilesize,
													width,
													height)
				end
			end
			
			self._static._tilesetCache[tilesetName] = self
			self._static._tilesetStrongRefs[tilesetName] = self
		end
	end
end

-- Getters --

function Tileset:getTileSize()
	return self._tilesize
end

function Tileset:getImage()
	return self._tilesetImage
end

function Tileset:_getQuad(tileIndex)
	if tileIndex > 0 and tileIndex <= self._numTiles then
		return self._quadList[tileIndex]
	end
end

-- Extra Methods --

function cleanTileCache()
	Tileset._static._tilesetStrongRefs = {}
end
