--[[
Copyright (c) 2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
ObjectList - A simple array for holding objects. Should provide fast iteration.

Currently has no methods for sorting the list. This may come at a later date.
--]]

require "lib.klutils.cabin"
require "lib.klutils.oops"

--[[
	The ObjectList class provides a facility for storing objects and then
	iterating through them in an efficient manner. Intended specifically for
	game use.
--]]
ObjectList = BaseObj:makeClass
{
	_data = nil,
	_indexMap = nil,
	_count = 0,
	_countIter = 0,
	_toRelease = nil,
	_releaseListEmpty = true,
	_parent = nil
}

--[[
	Object list constructor. An optional parent object may be specified, and
	should typically be set to the object that owns the list. Calls to
	insertAndOwn() will set the parent of any new members of the list to the
	object specified here.
--]]
function ObjectList:init(parentObj)
	self._data = {}
	self._indexMap = setmetatable({}, { __mode = "k" })
	self._toRelease = {}
	self._parent = parentObj
end

--[[
	Adds the object to (the end of) the list. The actual insertion location
	may change in a future implementation, so code should not rely on the list
	iterating through the objects in the order that they were inserted. Only
	one entry for a given object is allowed to exist inside the list. Returns
	false if nil is passed in or the object is already in the list. Otherwise
	returns true.
--]]
function ObjectList:insert(object)
	if not object or self._indexMap[object] then
		return false
	end
	
	self._count = self._count + 1
	
	self._data[self._count] = object
	self._indexMap[object] = self._count
	
	return true
end

--[[
	Adds the object to the list and takes ownership of the object by setting
	the object's parent to the list's parent. An object may only have one
	parent at any given time. Returns false if nil is passed in, the object
	already exists in the list, the list lacks a parent, or the object
	already has a parent. Otherwise
	returns true.
--]]
function ObjectList:insertAndOwn(object)
	if not object or self._indexMap[object] or not object.getParent
										or object:getParent() then
		return false
	elseif not self._parent then
		cabin.warning("An ObjectList without a parent tried to own an object. Not inserting object.")
		return false
	end
	
	self:insert(object)
	object:setParent(self._parent)
	
	return true
end

--[[
	Schedules the object for removal from the list. Note that the effect is not
	immediate - a call to sync() must be made to actually remove the object.
	Returns false if nil is passed in or if the object does not exist inside
	the list. Otherwise returns true.
--]]
function ObjectList:release(object)
	if not object then
		return false
	end
	
	local exists = false
	local index = self._indexMap[object]
	
	if index then
		self._toRelease[index] = true
		self._releaseListEmpty = false
		exists = true
	end
	
	return exists
end

--[[
	To provide consistent behavior, increase performance, and protect the data
	structure, changes to the entries in the list will only be visible after
	calling this function. Otherwise, "released" entries will still be
	considered part of the list, and newly added entries will not be iterated
	through. Generally, you will want to sync the list between each iteration
	of the list items. Never call this function during an iteration.
--]]
function ObjectList:sync()
	local oldCount = self._count
	
	-- Remove any objects marked for release.
	
	
	while not self._releaseListEmpty do
		-- We need to copy the objects to a local table because when we release
		--  the objects we set their parents to nil, which can call unparented()
		--  which could try to release more objects from the same list, hence
		--  the release list would be modified while it's still being iterated
		--  through.
		
		local releaseTable = {}
		
		for k, v in pairs(self._toRelease) do
			releaseTable[k] = v
		end
		
		self._toRelease = {}
		self._releaseListEmpty = true
		
		for k in pairs(releaseTable) do
			local object = self._data[k]
			
			if object.setParent and object:getParent() == self._parent then
				object:setParent(nil)
			end
			
			self._indexMap[object] = nil
			self._data[k] = nil
			
			self._count = self._count - 1
		end
	end
	
	-- Compact the array.
	
	local src, dest = 1, 1
	
	while src <= oldCount do
		if self._data[src] ~= nil then
			if src - dest > 0 then
				local relocatedObj = self._data[src]
				
				self._indexMap[relocatedObj] = dest
				self._data[dest] = relocatedObj
			end
			dest = dest + 1
		end
		src = src + 1
	end
	
	-- Nullify any trailing entries.
	
	while dest <= oldCount do
		self._data[dest] = nil
		
		dest = dest + 1
	end
	
	-- Update the iteration counter and clear the release table.
	
	self._countIter = self._count
end

--[[
	Provides Lua-style iterations for the list. When called, returns an
	iterator similar to that returned by pairs() or ipairs().
--]]
function ObjectList:iterate()
	local function iterator (list, i)
		i = i + 1
		
		if i < 1 or i > list._countIter then
			return nil
		else
			return i, list._data[i]
		end
	end
	
	return iterator, self, 0
end

--[[
	Returns the number of entries currently inside the list. This accounts for
	any objects that were inserted or deleted even if the list hasn't been
	synchronized.
--]]
function ObjectList:count()
	return self._count
end


--------------------------------------------------------------------------------


--[[
	The ListedObject class provides a small interface for objects that can be
	parented to an ObjectList.
--]]
ListedObj = BaseObj:makeClass
{
	_parent = nil
}

--[[
	Default constructor. Does nothing. Provided here only so that calling
	ListedObj.init(self) does not throw errors.
--]]
function ListedObj:init()
	
end


-- Returns the parent that this object belongs to.
function ListedObj:getParent()
	return self._parent
end

--[[
	Sets the parent of this object to the given object. Note that an object
	may only have one parent at a time.
--]]
function ListedObj:setParent(object)
	if self._parent then
		self:unparented(self._parent)
	end
	
	self._parent = object
	
	if object then
		self:parented(object)
	end
end

--[[
	Called after the object receives a new parent. To be overridden by the
	subclass.
--]]
function ListedObj:parented(parent)
	
end

--[[
	Called before the object loses its parent. To be overridden by the
	subclass.
--]]
function ListedObj:unparented(parent)
	
end
