--[[
Copyright (c) 2012-2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]


--[[
OOPS! A small Lua OOP lib.

A reasonably fast OOP system. Very simple. Fit for game use. Doesn't enforce
encapsulation, unfortunately.
--]]


-- Used to instantiate classes by string.
local _classNameTable = {}

-- Base object table
BaseObj = {}

--[[
	Copies the class member table from another class or forms a new one.
	Use this to define the member variables of a class, as well as the class
	you are subclassing from.
--]]
function BaseObj:makeClass(obj, name)

	local obj = obj or {}
	setmetatable(obj, self)
	self.__index = self
	
	if name and type(name) == 'string' then
		if _classNameTable[name] then
			error("A class was already defined with string name \"" .. name .. "\"")
		end
		
		_classNameTable[name] = obj
	end

	return obj

end

--[[
	Class object constructor. Define this for your class with any number of
	variables, and then call new() with matching parameters when you want a
	new object of that type. Call init on the superclass if you need to perform
	a construction on it as well:
	
	function Subclass:init(x, y, z)
		Superclass.init(self, x, y)
		
		-- Subclass initializations go here
	end
--]]
function BaseObj:init()
	
end

-- Constructs a new object of the specified class. Returns a reference to
--  the new object.
function BaseObj:new(...)
	
	local newObj = {}
	local arg = {...}
	
	setmetatable(newObj, self)
	self.__index = self

	newObj:init(unpack(arg))
	
	return newObj

end

-- Creates an instance of an object given its string name. Returns a
--  reference to the new object if the name is valid. Otherwise returns nil.
function instance(className, ...)
	if not _classNameTable[className] then
		return nil
	end
	
	return _classNameTable[className]:new(...)
end
