--[[
Copyright (c) 2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]

require "lib.klua"
require "lib.klutils.cabin"
require "lib.klutils.objectlist"

require "src.entities.puzzleboard"

require "src.entities.sliding-token"
require "src.entities.marble"
require "src.entities.paintball"

EditorBoard = PuzzleBoard:makeClass
{
	_tics = 0,
	_boardType = nil,
	_markerSize = 1.0
}

function EditorBoard:init(x, y, leveltable, levelnum, boardType, tileset)
	self._boardType = boardType
	
	self._tokenList = ObjectList:new()
	
	if not leveltable then
		cabin.error("Not a level pack.")
	elseif not leveltable[levelnum] then
		cabin.error("Level number does not exist in level pack.")
	end
	
	PuzzleBoard.init(self, x, y, leveltable, levelnum, boardType, tileset)
	
	if boardType == 'playerboard' then
		local level = leveltable[levelnum]
		
		if not level.tokens then
			cabin.error("Bad level data.")
		end
		
		for k, v in pairs(level.tokens) do
			if not v[1] or not v[2] or not v[3] then
				cabin.error("Bad token data.")
			end
			
			local class, tokenX, tokenY, properties = v[1], v[2], v[3], v[4]
			local token = instance(class, tokenX, tokenY, self, properties)
			
			self._tokenList:insert(token)
		end
		
		self._tokenList:sync()
	end
	
	-- Get scale for the marker.
	local tilegrid = self:getTilegrid()
	
	if not tilegrid then
		return nil
	end
	
	local tileset = tilegrid:getTileset()
	
	if not tileset then
		return nil
	end
	
	self._markerSize = tileset:getTileSize() / 48
end

function EditorBoard:parented(parent)
	if parent.addEntity and self._tokenList then
		for i, v in self._tokenList:iterate() do
			parent:addEntity(v)
		end
	end
end

function EditorBoard:unparented(parent)
	for i, v in self._tokenList:iterate() do
		v:destroy()
	end
end

function EditorBoard:think()
	-- Propagate visible status
	for i, v in ipairs(self._tokenList) do
		v:setVisible(self:isVisible())
	end
end

function EditorBoard:doObjectPlace(boardX, boardY)
	local grid = self:getTilegrid()
		
	if not grid then
		return
	end
	
	if not input.held('edit-entity') then
		if input.held('edit-white') then
			grid:setTileValue(boardX, boardY, tileIndex['floorWhite'])
		elseif input.held('edit-red') then
			grid:setTileValue(boardX, boardY, tileIndex['floorRed'])
		elseif input.held('edit-yellow') then
			grid:setTileValue(boardX, boardY, tileIndex['floorYellow'])
		elseif input.held('edit-blue') then
			grid:setTileValue(boardX, boardY, tileIndex['floorBlue'])
		elseif input.held('edit-orange') then
			grid:setTileValue(boardX, boardY, tileIndex['floorOrange'])
		elseif input.held('edit-green') then
			grid:setTileValue(boardX, boardY, tileIndex['floorGreen'])
		elseif input.held('edit-purple') then
			grid:setTileValue(boardX, boardY, tileIndex['floorPurple'])
		elseif input.held('edit-wall') then
			grid:setTileValue(boardX, boardY, tileIndex['wallWhite'])
		end
	elseif self._boardType ~= 'goal' then
		local color = nil
		
		if input.downOnce('edit-red') then
			color = 'red'
		elseif input.downOnce('edit-yellow') then
			color = 'yellow'
		elseif input.downOnce('edit-blue') then
			color = 'blue'
		elseif input.downOnce('edit-orange') then
			color = 'orange'
		elseif input.downOnce('edit-green') then
			color = 'green'
		elseif input.downOnce('edit-purple') then
			color = 'purple'
		end
		
		if color then
			local paintBall = Paintball:new(boardX, boardY, self, color)
			
			local parent = self:getParent()
			
			if parent then
				parent:addEntity(paintBall)
				
				self._tokenList:insert(paintBall)
			end
		end
	end
	
	if input.downOnce('edit-delete') then
		for i, v in self._tokenList:iterate() do
			local tokenBoardX, tokenBoardY = v:getBoardPosition()
			
			if tokenBoardX == boardX and tokenBoardY == boardY then
				self._tokenList:release(v)
				
				v:destroy()
			end
		end
	end
	
	self._tokenList:sync()
end

function EditorBoard:getMarkerSize()
	return self._markerSize
end

function EditorBoard:dataToStdout()
	-- There are better ways of saving data, but this is 1GAM. I need
	--  something quick and dirty.
	
	local grid = self:getTilegrid()
	
	if not grid then
		return
	end
	
	print(self._boardType .. " = {")
		
	local width, height = grid:getSize()
	
	for y = 0, height - 1 do
		local row = "\t{ "
			
		for x = 0, width - 1 do
			if x > 0 then
				row = row .. ", "
			end
			
			local value = grid:getTileValue(x, y)
			-- For sake of padding
			local valueString = ""
			
			if value >= 10 then
				valueString = value
			else
				valueString = " " .. value
			end
			
			row = row .. valueString
		end
		
		print(row .. " },")
	end
	
	print("},")
	
	if self._tokenList:count() > 0 then
		print("tokens = {")
		
		for i, v in self._tokenList:iterate() do
			local x, y = v:getBoardPosition()
			
			print("\t{ \"Paintball\", " .. x .. ", " .. y .. ", \""
								.. v:getPaintColor() .. "\" },")
		end
		
		print("},")
	end
end
