--[[
Copyright (c) 2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]

require "lib.klua"
require "lib.klutils.cabin"
require "lib.klutils.objectlist"

require "src.entities.puzzleboard"

require "src.entities.sliding-token"
require "src.entities.marble"
require "src.entities.paintball"

require "src.tileconstants"

local _floorStartIndex = 1
local _wallStartIndex = 15

local _destOffsetTable = 
{
	['left'] = {x = -1, y = 0},
	['right'] = {x = 1, y = 0},
	['up'] = {x = 0, y = -1},
	['down'] = {x = 0, y = 1}
}

PlayerBoard = PuzzleBoard:makeClass
{
	_tics = 0,
	_fadeTics = 0,
	_targetScale = 1.0,
	_targetX = 0,
	_targetY = 0,
	_targetTween = 30,
	_bumpAnimTics = -1,
	_lastBumpTic = 0,
	_lastSplatSoundTic = 0,
	_enableControls = true
}

function PlayerBoard:init(x, y, leveltable, levelnum)
	self:setScale(0, 0)
	
	self._targetX = x
	self._targetY = y
	
	self._tokenList = ObjectList:new()
	self._unavailableColors = {
		['white'] = true,
		['red'] = true,
		['yellow'] = true,
		['blue'] = true,
		['orange'] = true,
		['green'] = true,
		['purple'] = true,
	}
	
	self._bumpSound = Sound:new("bump.ogg")
	self._paintSound = Sound:new({"paint1.ogg", "paint2.ogg", "paint3.ogg"})
	
	if not leveltable then
		cabin.error("Not a level pack.")
	elseif not leveltable[levelnum] then
		cabin.error("Level number does not exist in level pack.")
	end
	
	PuzzleBoard.init(self, x, y, leveltable, levelnum, 'playerboard', 'board')
	
	local level = leveltable[levelnum]
	
	if not level.tokens then
		cabin.error("Bad level data.")
	end
	
	for k, v in pairs(level.tokens) do
		if not v[1] or not v[2] or not v[3] then
			cabin.error("Bad token data.")
		end
		
		local class, tokenX, tokenY, properties = v[1], v[2], v[3], v[4]
		local token = instance(class, tokenX, tokenY, self, properties)
		token:doTransition()
		
		self._tokenList:insert(token)
		
		-- We have a paintball of this color, so remove it from deadlock
		--  consideration
		if self._unavailableColors[properties] then
			self._unavailableColors[properties] = false
		end
	end
	
	self._tokenList:sync()
end

function PlayerBoard:parented(parent)
	if parent.addEntity and self._tokenList then
		for i, v in self._tokenList:iterate() do
			parent:addEntity(v)
		end
	end
end

function PlayerBoard:unparented(parent)
	for i, v in self._tokenList:iterate() do
		v:destroy()
	end
end

function PlayerBoard:setControlsEnabled(areEnabled)
	self._enableControls = (areEnabled == true)
end

function PlayerBoard:_tiltControl()
	if not self:isTilting() then
		local dir = nil
		local left, right, up, down = input.held('left'),
									input.held('right'),
									input.held('up'),
									input.held('down')
		
		if left and not right then
			dir = 'left'
		elseif right and not left then
			dir = 'right'
		elseif up and not down then
			dir = 'up'
		elseif down and not up then
			dir = 'down'
		end
		
		-- Limit the number of times you can tilt a second unless tilting
		--  in a different direction
		if dir and (dir ~= self._lastTiltDirection
					or (self._tics - self._lastBumpTic) > 30) then
			self:tilt(dir)
		end
	else
		local tilegrid = self:getTilegrid()
	
		if not tilegrid then
			return
		end
		
		local sizeW, sizeH = tilegrid:getSize()
		
		if self._tilterCol then
			if self._lastTiltDirection == 'left' then
				self._tilterCol = self._tilterCol + 1
				
				if self._tilterCol >= sizeW then
					self._tilterCol = nil
				end
			elseif self._lastTiltDirection == 'right' then
				self._tilterCol = self._tilterCol - 1
				
				if self._tilterCol < 0 then
					self._tilterCol = nil
				end
			end
		elseif self._tilterRow then
			if self._lastTiltDirection == 'up' then
				self._tilterRow = self._tilterRow + 1
				
				if self._tilterRow >= sizeH then
					self._tilterRow = nil
				end
			elseif self._lastTiltDirection == 'down' then
				self._tilterRow = self._tilterRow - 1
				
				if self._tilterRow < 0 then
					self._tilterRow = nil
				end
			end
		end
		
		self:_tiltTokenColumnRow()
	end
end

function PlayerBoard:doDisappear()
	-- Ugh, I need an actual tweening API, but I don't have time for that.
	--  So umm... just make the balls fly off the screen using the existing
	--  code.
	
	self._targetTween = 60
	self._fadeTics = 30
	self._enableControls = false
	
	for i, v in self._tokenList:iterate() do
		v:doDisappear()
	end
end

function PlayerBoard:think()
	self._tics = self._tics + 1
	self._fadeTics = self._fadeTics + 1
	
	-- Initial transition
	if self._fadeTics <= self._targetTween then
		local tween = math.sin(self._fadeTics / 60.0 * math.pi)
		self:setScale(tween * self._targetScale, tween * self._targetScale)
		self:setAngle(1 - (tween))
	elseif self._enableControls then
		self:_tiltControl()
	end
	
	-- Tilt animation
	if self._bumpAnimTics >= 0 then
		self._bumpAnimTics = self._bumpAnimTics + 0.2
		
		local offset = _destOffsetTable[self._lastTiltDirection]
		
		if offset then
			local bumpX = self._targetX
						+ offset.x * math.sin(self._bumpAnimTics) * 20
			local bumpY = self._targetY
						+ offset.y * math.sin(self._bumpAnimTics) * 20
			
			self:setPosition(bumpX, bumpY)
		end
		
		if self._bumpAnimTics > math.pi then
			self._bumpAnimTics = -1
			
			self:setPosition(self._targetX, self._targetY)
		end
	end
	
	-- Propagate visible status
	for i, v in self._tokenList:iterate() do
		v:setVisible(self:isVisible())
	end
end

function PlayerBoard:boardToScreenCoords(boardX, boardY)
	
	local originX, originY = self:getPosition()
	local sizeX, sizeY = self:getPosition()
	
	local tilegrid = self:getTilegrid()
	
	if not tilegrid then
		return nil
	end
	
	local sizeW, sizeH = tilegrid:getSize()
	
	local tileset = tilegrid:getTileset()
	
	if not tileset then
		return nil
	end
	
	local tilesize = tileset:getTileSize()
	
	sizeW = sizeW * tilesize
	sizeH = sizeH * tilesize
	
	originX = originX - sizeW / 2 + (boardX + 0.5) * tilesize
	originY = originY - sizeH / 2 + (boardY + 0.5) * tilesize
	
	return originX, originY
end

function PlayerBoard:isTilting()
	local tilting = false
	
	if self._tilterRow or self._tilterCol then
		tilting = true
	else
		for i, v in self._tokenList:iterate() do
			if v:isMoving() then
				tilting = true
				break
			end
		end
	end
	
	return tilting
end

function PlayerBoard:isFree(boardX, boardY)
	local free = true
	
	local tilegrid = self:getTilegrid()
	
	if tilegrid then
		local tileValue = tilegrid:getTileValue(boardX, boardY)
		
		if not tileValue or tileValue < tileIndex['floorStart']
					or tileValue >= tileIndex['wallStart'] then
			free = false
		else
			for i, v in self._tokenList:iterate() do
				local posX, posY = v:getBoardPosition()
				
				if posX == boardX and posY == boardY then
					free = false
					break
				end
			end
		end
	end
	
	return free
end

function PlayerBoard:tilt(direction)
	
	self._lastBumpTic = self._tics
	self._bumpSound:play()
	
	local minX, minY, maxX, maxY = 9999, 9999, -1, -1
	
	for i, v in self._tokenList:iterate() do
		local boardX, boardY = v:getBoardPosition()
		
		minX = (boardX <= minX) and boardX or minX
		minY = (boardY <= minY) and boardY or minY
		maxX = (boardX >= maxX) and boardX or maxX
		maxY = (boardY >= maxY) and boardY or maxY
	end
	
	if direction == 'left' then
		self._tilterCol = minX - 1
	elseif direction == 'right' then
		self._tilterCol = maxX + 1
	elseif direction == 'up' then
		self._tilterRow = minY - 1
	elseif direction == 'down' then
		self._tilterRow = maxY + 1
	end
	
	self._lastTiltDirection = direction
	self._bumpAnimTics = 0
end

function PlayerBoard:tileCrossed(boardX, boardY, crosserObj)
	local tilegrid = self:getTilegrid()
	local targetTile = tileIndex['floorWhite'] + colorToOffset[crosserObj:getPaintColor()]
	
	if tilegrid then
		local value = tilegrid:getTileValue(boardX, boardY)
		
		if targetTile ~= value then
			-- Don't play the sound TOO often
			if (self._tics - self._lastSplatSoundTic) > 4 then
				self._lastSplatSoundTic = self._tics
				
				self._paintSound:play()
			end
			
			tilegrid:setTileValue(boardX, boardY, targetTile)
		end
	end
end

function PlayerBoard:_tiltTokenColumnRow()
	local dir = self._lastTiltDirection
	local axis = ''
	
	if (dir == 'left' or dir == 'right') then
		axis = 'horizontal'
		
		if not self._tilterCol then
			return
		end
	elseif (dir == 'up' or dir == 'down') then
		axis = 'vertical'
		
		if not self._tilterRow then
			return
		end
	end
	
	for i, v in self._tokenList:iterate() do
		local posX, posY = v:getBoardPosition()
		
		if (axis == 'horizontal' and posX == self._tilterCol)
				or (axis == 'vertical' and posY == self._tilterRow) then
			self:_tiltToken(v, posX, posY, dir)
		end
	end
end

function PlayerBoard:_tiltToken(token, originX, originY, direction)
	
	local tilegrid = self:getTilegrid()
	
	if not tilegrid then
		return
	end
	
	local sizeW, sizeH = tilegrid:getSize()
	local offset = _destOffsetTable[direction]
	
	if offset then
		local destX = originX
		local destY = originY
		
		while destX >= 0 and destX < sizeW and destY >= 0 and destY < sizeH do
			if not self:isFree(destX + offset.x, destY + offset.y) then
				break
			end
			
			destX = destX + offset.x
			destY = destY + offset.y
		end
		
		token:rollToBoardPosition(destX, destY)
	end
end

function PlayerBoard:paintballColorAvailable(color)
	return (self._unavailableColors[color] ~= true)
end
