--[[
Copyright (c) 2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]

require "lib.klua"
require "lib.klutils.cabin"

local abs, floor = math.abs, math.floor

SlidingToken = Entity:makeClass
{
	_accel = 0.02,
	_isMoving = false,
	_animVelX = 0,
	_animVelY = 0,
	_tics = 0
}

function SlidingToken:init(boardX, boardY, boardObj)
	if not boardObj then
		cabin.error("SlidingToken must belong to a board object.")
	end
	
	self._boardX = boardX
	self._boardY = boardY
	self._targetX = boardX
	self._targetY = boardY
	self._animPosX = boardX
	self._animPosY = boardY
	
	self._boardObj = boardObj
	
	local originX, originY = boardObj:boardToScreenCoords(boardX, boardY)
	
	Entity.init(self, originX, originY)
end

function SlidingToken:doTransition()
	self:setScale(0, 0)
	self._tics = -40
end

function SlidingToken:doDisappear()
	-- Ugh, I need an actual tweening API, but I don't have time for that.
	--  So umm... just make the balls fly off the screen using the existing
	--  code.
	
	self._targetY = -9999
	self._finishPlaying = true
end

function SlidingToken:_animateTilt()
	local oldX = self._animPosX
	local oldY = self._animPosY
	
	if self._animPosX + self._animVelX < self._targetX then
		-- At or past target
		if self._animVelX < 0 then
			self._animVelX = 0
			self._animPosX = self._targetX
		else
			self._animVelX = self._animVelX + self._accel
		end
	elseif self._animPosX + self._animVelX > self._targetX then
		-- At or past target
		if self._animVelX > 0 then
			self._animVelX = 0
			self._animPosX = self._targetX
		else
			self._animVelX = self._animVelX - self._accel
		end
	end
	
	if self._animPosY + self._animVelY < self._targetY then
		-- At or past target
		if self._animVelY < 0 then
			self._animVelY = 0
			self._animPosY = self._targetY
		else
			self._animVelY = self._animVelY + self._accel
		end
	elseif self._animPosY + self._animVelY > self._targetY then
		-- At or past target
		if self._animVelY > 0 then
			self._animVelY = 0
			self._animPosY = self._targetY
		else
			self._animVelY = self._animVelY - self._accel
		end
	end
	
	self._isMoving = (self._animVelX ~= 0 or self._animVelY ~= 0)
	
	self._animPosX = self._animPosX + self._animVelX
	self._animPosY = self._animPosY + self._animVelY
	
	if self._boardObj then
		local screenX, screenY = self._boardObj:boardToScreenCoords(self._animPosX,
															self._animPosY)
		self:setPosition(screenX, screenY)
		
		local prevTileX = floor(oldX + 0.5)
		local prevTileY = floor(oldY + 0.5)
		local nextTileX = floor(self._animPosX + 0.5)
		local nextTileY = floor(self._animPosY + 0.5)
		
		if not self._finishPlaying then
			if prevTileX < nextTileX then
				for x = prevTileX + 1, nextTileX do
					self._boardObj:tileCrossed(x, prevTileY, self)
				end
			elseif prevTileX > nextTileX then
				for x = nextTileX, prevTileX - 1 do
					self._boardObj:tileCrossed(x, prevTileY, self)
				end
			elseif prevTileY < nextTileY then
				for y = prevTileY + 1, nextTileY do
					self._boardObj:tileCrossed(prevTileX, y, self)
				end
			elseif prevTileY > nextTileY then
				for y = nextTileY, prevTileY - 1 do
					self._boardObj:tileCrossed(prevTileX, y, self)
				end
			end
		end
	end
end

function SlidingToken:think()
	self._tics = self._tics + 1
	
	self:_animateTilt()
	
	local targetScaleX = 1.0 + math.abs(self._animVelX * 0.5)
							- math.abs(self._animVelY * 0.5)
	local targetScaleY = 1.0 / targetScaleX
	
	-- Appearing animation
	if self._tics > -10 and self._tics < 0 then
		local tween = math.sin((self._tics + 10) * math.pi * 0.05)
		self:setScale(tween * targetScaleX, tween * targetScaleY)
	elseif self._tics >= 0 then
		self:setScale(targetScaleX, targetScaleY)
	end
end

function SlidingToken:getBoardPosition()
	return self._boardX, self._boardY
end

function SlidingToken:rollToBoardPosition(boardX, boardY)
	if not self._boardObj or self._finishPlaying then
		cabin.error("SlidingToken must belong to a board object.")
	end
	
	self._boardX = boardX
	self._boardY = boardY
	self._targetX = boardX
	self._targetY = boardY
end

function SlidingToken:isMoving()
	return self._isMoving
end

function SlidingToken:getPaint()
	return 0
end

function SlidingToken:getPaintColor()
	return ""
end
