--[[
Copyright (c) 2013 Daniel Hagerstrand

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software in a
product, an acknowledgment in the product documentation would be appreciated
but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
--]]

require "lib.klua"
require "lib.klutils.cabin"

require "src.setmode"

require "src.entities.playerboard"
require "src.entities.goalboard"
require "src.entities.bgball"
require "src.entities.mistake"
require "src.entities.check"

require "src.entities.editor.editorboard"
require "src.entities.editor.marker"

require "src.tileconstants"

local _versionString = "v1.0 - Initial Release"

PlayPuzzleScene = Scene:makeClass
{
	_currentMenu = 'select',
	_currentLevel = -1,
	_menuSelection = 0,
	_menuPrevSelection = 0,
	_menuAnimationTics = 4,
	_numLevels = 0,
	_newPossibleSolution = false,
	_finishTimer = -1,
	_firstUnfinished = 1
}

function PlayPuzzleScene:init()
	
	Scene.init(self)
	
	self:setName("Play Puzzle")
	-- We have double the height so we can "scroll" between the menu and
	--  the game.
	self:setAreaSize(1920, 2160)
	
	self._leveltable = propertyfile.get("levelpacks/standard")
	
	if not self._leveltable then
		cabin.error("Levelpack data is bad.")
	end
	
	self._solvedLevels = {}
	self._menuObjs = {}
	self._boardObjs = {}
	
	for i = 1, 5 do
		local ball = BgBall:new(-1000, -1000)
		self:addEntity(ball)
	end
	
	self:setBackgroundColor(100, 173, 230)
	
	-- Menu/game controls
	input.setControl('select', 'return')
	input.setControl('back', 'escape')
	input.setControl('reset', 'r')
	input.setControl('fullscreen', 'f')
	input.setControl('skip-right', 'pagedown')
	input.setControl('skip-left', 'pageup')
	
	-- Editor controls
	input.setControl('edit', 'e')
	input.setControl('edit-white', '0')
	input.setControl('edit-red', '1')
	input.setControl('edit-yellow', '2')
	input.setControl('edit-blue', '3')
	input.setControl('edit-orange', '4')
	input.setControl('edit-green', '5')
	input.setControl('edit-purple', '6')
	input.setControl('edit-entity', {'lctrl', 'rctrl'})
	input.setControl('edit-wall', '-')
	input.setControl('edit-delete', 'delete')
	input.setControl('edit-switch', 'tab')
	input.setControl('edit-save', 's')
	
	self._menuObjs.title = TextEntity:new(0, 150, "Untilted Game", "engebrechtre ex", 200)
	self._menuObjs.title:setBounds(1920, 'center')
	self._menuObjs.subtitle = TextEntity:new(0, 950, "Game by \"Dattorz\" (@gt_effbee) - Music by Ryan Melmoth - For additional credits, see CREDITS-SOUND"
								.. "\n\nPress F to toggle Fullscreen.",
										"engebrechtre ex bd", 32)
	self._menuObjs.subtitle:setBounds(1920, 'center')
	self._menuObjs.version = TextEntity:new(0, 400, _versionString, "engebrechtre ex", 48)
	self._menuObjs.version:setBounds(1920, 'center')
	self._menuObjs.levelTitle = TextEntity:new(0, 800, "", "engebrechtre ex", 48)
	self._menuObjs.levelTitle:setBounds(1920, 'center')
	
	self:addEntity(self._menuObjs.title)
	self:addEntity(self._menuObjs.version)
	self:addEntity(self._menuObjs.subtitle)
	self:addEntity(self._menuObjs.levelTitle)
	
	self:_createLevelSelect()
	
	music.setSong("trashcanalley.ogg")
	
	self._oopsSound = Sound:new("oops.ogg")
	self._finishSound = Sound:new("finish.ogg")
	
	-- For initialization purposes
	self:_switchMenuMode('select')
	self:_changeMenuSelection(self._firstUnfinished)
end

function PlayPuzzleScene:_updateSave()
	local solvedLines = ""
	
	for k, v in pairs(self._solvedLevels) do
		if v then
			solvedLines = solvedLines .. tostring(k) .. '\n'
		end
	end
	
	local returnCode, reason = pcall(love.filesystem.write, 'solved', solvedLines)
	
	if not returnCode then
		cabin.warning("Could not save progress: " .. tostring(reason))
	end
end

function PlayPuzzleScene:_createLevelSelect()
	local i = 1
	
	while self._leveltable[i] do
		-- Is there a more efficient method than creating a separate object for
		--  every level? Yes, but this is 1GAM.
		local level = PuzzleBoard:new(-1000, -1000, self._leveltable, i, 'goal', 'board-select')
		local check = Check:new(-1000, -1000)
		
		self._menuObjs['thumb' .. i] = level
		self._menuObjs['check' .. i] = check
		self:addEntity(level)
		self:addEntity(check)
		
		self._numLevels = i
		i = i + 1
	end
	
	-- Load save data
	if love.filesystem.exists('solved') then
		for line in love.filesystem.lines('solved') do
			local num = tonumber(line)
			if num then
				self._solvedLevels[num] = true
			end
		end
	end
	
	-- Find lowest unsolved
	for i = 1, self._numLevels do
		if self._solvedLevels[i] then
			self._firstUnfinished = i + 1
		else
			break
		end
	end
end

function PlayPuzzleScene:_createPuzzle(mode)
	if self._boardObjs.main then
		for k, v in pairs(self._boardObjs) do
			v:destroy()
		end
		
		self._boardObjs = {}
	end
	
	if not mode or mode == 'game' then
		self._boardObjs.main = PlayerBoard:new(640, 1620, self._leveltable, self._currentLevel)
		self._boardObjs.goal = GoalBoard:new(1568, 1396, self._leveltable, self._currentLevel)
	elseif mode == 'edit' then
		self._boardObjs.main = EditorBoard:new(640, 1620, self._leveltable, self._currentLevel, 'playerboard', 'board')
		self._boardObjs.goal = EditorBoard:new(1568, 1396, self._leveltable, self._currentLevel, 'goal', 'board-mini')
	end
	
	if self._leveltable[self._currentLevel].name then
		self._boardObjs.title = TextEntity:new(1264, 1652, self._currentLevel
						.. ". " .. self._leveltable[self._currentLevel].name,
						"engebrechtre ex", 48)
		self._boardObjs.title:setBounds(600, 'center')
	end
	
	self:addEntity(self._boardObjs.main)
	self:addEntity(self._boardObjs.goal)
	self:addEntity(self._boardObjs.title)
end

function PlayPuzzleScene:_changeMenuSelection(target)
	target = (target >= 1) and target or 1
	target = (target <= self._numLevels) and target or self._numLevels
	
	if target ~= self._menuSelection then
		self._menuPrevSelection = self._menuSelection
		self._menuSelection = target
		self._menuAnimationTics = 0
		
		if self._leveltable and self._leveltable[target]
						and self._leveltable[target].name then
			self._menuObjs.levelTitle:setString(target .. ". " .. self._leveltable[target].name)
		end
	end
end

function PlayPuzzleScene:_menuThink()
	if self._menuAnimationTics >= math.pi / 2 then
		self._menuAnimationPosX = self._menuSelection
		
		if input.held('right') then
			self:_changeMenuSelection(self._menuSelection + 1)
		elseif input.held('left') then
			self:_changeMenuSelection(self._menuSelection - 1)
		elseif input.held('skip-right') then
			self:_changeMenuSelection(self._menuSelection + 10)
		elseif input.held('skip-left') then
			self:_changeMenuSelection(self._menuSelection - 10)
		end
	else
		self._menuAnimationTics = self._menuAnimationTics + 0.15
		
		local offset = self._menuSelection - self._menuPrevSelection
		
		self._menuAnimationPosX = self._menuPrevSelection + offset * math.sin(self._menuAnimationTics)
	end
	
	for i = 1, self._numLevels do
		local visible = i >= self._menuAnimationPosX - 3 and i <= self._menuAnimationPosX + 3
		local alpha = math.max(0, 255 - math.abs(i - self._menuAnimationPosX) * 96)
		local position = 960 + (i - self._menuAnimationPosX) * 300
		
		-- OMG, what a hack.
		--  Next time I'll be sure to make a class that groups entities
		--  together so that I can avoid stupid things like this.
		
		self._menuObjs['thumb' .. i]:setVisible(visible)
		self._menuObjs['thumb' .. i]:setColor(255, 255, 255, alpha)
		self._menuObjs['thumb' .. i]:setPosition(position, 650)
		
		self._menuObjs['check' .. i]:setVisible(visible
								and self._solvedLevels[i])
		self._menuObjs['check' .. i]:setColor(255, 255, 255, alpha)
		self._menuObjs['check' .. i]:setPosition(position, 650)
	end
	
	if input.downOnce('select') then
		if self._currentLevel ~= self._menuSelection then
			self._currentLevel = self._menuSelection
			self:_createPuzzle()
		end
		self:_switchMenuMode('puzzle')
	elseif input.downOnce('edit') and gameloop.getDebugEnabled() then
		self._currentLevel = self._menuSelection
		self:_createPuzzle('edit')
		self:_switchMenuMode('edit')
	end
end

function PlayPuzzleScene:_checkMatching(gridA, gridB)
	local sizeW, sizeH = gridB:getSize()
	
	for y = 0, sizeH - 1 do
		for x = 0, sizeW - 1 do
			if gridA:getTileValue(x, y) ~= gridB:getTileValue(x, y) then
				return false
			end
		end
	end
	
	return true
end

function PlayPuzzleScene:_checkDeadlock()
	local isDeadlocked = false
	
	if not self._boardObjs.main or not self._boardObjs.goal then
		return
	end
	
	local mainGrid = self._boardObjs.main:getTilegrid()
	local goalGrid = self._boardObjs.goal:getTilegrid()
	
	if not mainGrid or not goalGrid then
		return
	end
	
	local sizeW, sizeH = goalGrid:getSize()
	local mistakeList = {}
	local mistakeIndex = 1
	
	for y = 0, sizeH - 1 do
		for x = 0, sizeW - 1 do
			local curValue = mainGrid:getTileValue(x, y)
			local targetValue = goalGrid:getTileValue(x, y)
			local color = tileIndexToColor[targetValue]
			local available = self._boardObjs.main:paintballColorAvailable(color)
			
			if curValue ~= targetValue and not available then
				isDeadlocked = true
				
				mistakeList[mistakeIndex] = {}
				mistakeList[mistakeIndex].x = x
				mistakeList[mistakeIndex].y = y
				mistakeIndex = mistakeIndex + 1
			end
		end
	end
	
	return isDeadlocked, mistakeList
end

function PlayPuzzleScene:_puzzleThink()
	if not self._boardObjs.main or not self._boardObjs.goal then
		return
	end
	
	if self._boardObjs.main:isTilting() then
		self._newPossibleSolution = true
	else
		if self._newPossibleSolution then
			-- Only check the solution once per turn.
			self._newPossibleSolution = false
			
			local playerGrid = self._boardObjs.main:getTilegrid()
			local goalGrid = self._boardObjs.goal:getTilegrid()
				
			if not playerGrid or not goalGrid then
				return
			end
			
			if self:_checkMatching(playerGrid, goalGrid)
						and self._finishTimer < 0 then
				
				if not self._solvedLevels[self._currentLevel] then
					self._solvedLevels[self._currentLevel] = true
					self:_changeMenuSelection(self._currentLevel + 1)
				end
				
				-- Start the finish animation
				self._finishTimer = 0
				self._boardObjs.main:doDisappear()
				self._boardObjs.goal:doDisappear()
				
				self._boardObjs.title:destroy()
				self._boardObjs.title = nil
				
				self._boardObjs.win = TextEntity:new(0, 0,
									"Complete!",
									"engebrechtre ex", 128)
				self._boardObjs.win:setBounds(1920, 'center')
				self:addEntity(self._boardObjs.win)
				
				self._finishSound:play()
			else
				local deadlocked, mistakeList = self:_checkDeadlock()
				if deadlocked and not self._boardObjs.retry then
					self._boardObjs.retry = TextEntity:new(1264, 1780,
									"You probably want to press 'R' to retry.",
									"engebrechtre ex", 48)
					self._boardObjs.retry:setBounds(600, 'center')
					self:addEntity(self._boardObjs.retry)
					
					for i, v in ipairs(mistakeList) do
						local mistakeName = 'mistake-' .. i
						
						self._boardObjs[mistakeName] = Mistake:new(v.x, v.y,
												self._boardObjs.main)
						self:addEntity(self._boardObjs[mistakeName])
					end
					
					self._oopsSound:play()
				end
			end
		end
	end
	
	-- Finish animation
	if self._finishTimer >= 0 and self._boardObjs.win then
		self._finishTimer = self._finishTimer + 1
		
		self._boardObjs.win:setPosition(0, 1620
							+ math.sin(self._finishTimer * 0.2) * 3000
							/ (self._finishTimer))
		
		if self._finishTimer >= 90 then
			self:_updateSave()
			
			-- Don't "resume" the level when reselecting it.
			self._currentLevel = -1
			
			self._finishTimer = -1
			self:_switchMenuMode('select')
		end
	else
		-- Don't allow switching if the finish animation is playing
		if input.downOnce('back') then
			self:_switchMenuMode('select')
		elseif input.downOnce('reset') then
			self:_createPuzzle()
		end
	end
end

function PlayPuzzleScene:_editThink()
	if not self._boardObjs.marker then
		self._boardObjs.marker = Marker:new(0, 0, self._boardObjs.main)
		self:addEntity(self._boardObjs.marker)
	end
	
	local marker = self._boardObjs.marker
	
	if input.downOnce('edit-switch') then
		local board = marker:getParentBoard()
		
		if board == self._boardObjs.main then
			marker:setParentBoard(self._boardObjs.goal)
		else
			marker:setParentBoard(self._boardObjs.main)
		end
	end
	
	if input.downOnce('edit-save') then
		self._boardObjs.main:dataToStdout()
		self._boardObjs.goal:dataToStdout()
	end
end

function PlayPuzzleScene:_switchMenuMode(newMode)
	if newMode == 'select' then
		self:centerScrollVector(960, 540)
		
		if self._currentMenu == 'puzzle' and self._boardObjs.main then
			self._boardObjs.main:setControlsEnabled(false)
		end
		
		for k, v in pairs(self._menuObjs) do
			v:setVisible(true)
		end
		for k, v in pairs(self._boardObjs) do
			v:setVisible(false)
		end
	else
		self:centerScrollVector(960, 1620)
		
		if newMode == 'puzzle' then
			self._boardObjs.main:setControlsEnabled(true)
		end
		
		for k, v in pairs(self._menuObjs) do
			v:setVisible(false)
		end
		for k, v in pairs(self._boardObjs) do
			v:setVisible(true)
		end
	end
	
	self._currentMenu = newMode
end

function PlayPuzzleScene:prestep()
	if self._currentMenu == 'select' then
		self:_menuThink()
	elseif self._currentMenu == 'puzzle' then
		self:_puzzleThink()
	elseif self._currentMenu == 'edit' then
		self:_editThink()
	end
	
	-- This does not belong here. Ideally it should go in some sort of global
	--  key binding manager inside the framework but it's easier to put it here.
	--  Once again my excuse is that this is 1GAM.
	if input.downOnce('fullscreen') then
		toggleMode()
	end
end

